package ru.dbotthepony.kstarbound.world.entities.api

import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import ru.dbotthepony.kstarbound.lua.LuaThread

interface ScriptedEntity {
	val lua: LuaThread

	// Call a script function directly with the given arguments, should return
	// nothing only on failure.
	fun callScript(fnName: String, arguments: JsonArray): JsonElement {
		return lua.invokeGlobal(fnName, 1, {
			ensureExtraCapacity(arguments.size() + 4)

			for (argument in arguments) {
				push(argument)
			}

			arguments.size()
		}, { getJson() ?: JsonNull.INSTANCE }).orElse(JsonNull.INSTANCE)
	}

	// Execute the given code directly in the underlying context, return nothing
	// on failure.
	fun evalScript(code: String): JsonElement {
		return lua.eval(code)
	}
}
